def removesuffix(self: str, suffix: str) -> str:
    # suffix='' should not call self[:-0].
    if suffix and self.endswith(suffix):
        return self[: -len(suffix)]
    else:
        return self[:]



from .file_editing import change_key_name
import argparse
import glob

parser = argparse.ArgumentParser(
    description="Porting JSON file(s) to Cluster utils >=3.0"
)
parser.add_argument(dest="files", type=str, help="Input file(s)", nargs="+")


if __name__ == "__main__":
    args = parser.parse_args()

    for setting_file in args.files:
        change_key_name(
            setting_file, prefixes=[], old_name="model_dir", new_name="working_dir"
        )
        change_key_name(
            setting_file, prefixes=[], old_name="default_json", new_name="__import__"
        )
        change_key_name(
            setting_file,
            prefixes=["fixed_params"],
            old_name="default_json",
            new_name="__import_promise__",
        )



import glob
import json


def get_all_jsons():
    return glob.glob(f"settings/**/*.json", recursive=True)


def change_key_name(setting_file, prefixes, old_name, new_name, conditions=None):
    conditions = conditions or {}

    with open(setting_file) as f:
        try:
            orig_dict = json.load(f)
        except Exception as e:
            print(f"{setting_file}: loading failed with error {e}... skipping")
            return False
    if not check_correct_conditions(orig_dict, conditions):
        print(f"{setting_file}: Conditions {conditions} not met ... skipping")
        return False
    as_dict = orig_dict
    try:
        for dict_key in prefixes:
            as_dict = as_dict[dict_key]
    except KeyError:
        print(f"{setting_file}: Not all prefixes {prefixes} present ... skipping")
        return False  # Prefixes not present in JSON file

    if old_name in as_dict:
        old_val = as_dict[old_name]
        as_dict[new_name] = old_val
        print(
            f"{setting_file}: Key {'.'.join(prefixes+[old_name])} renamed to {'.'.join(prefixes+[new_name])}"
        )
        del as_dict[old_name]
        with open(setting_file, "w") as f:
            json.dump(orig_dict, f, indent=4)
        return True

    return False


def check_correct_conditions(dct, conditions):
    for lhs, rhs in conditions.items():
        if not rhs == dct.get(
            lhs, float("nan")
        ):  # using that 'nan' is not equal to ANYTHING in Python
            return False
    return True


def add_key(
    setting_file, prefixes, new_key, default_value, override=False, conditions=None
):
    conditions = conditions or {}

    with open(setting_file) as f:
        try:
            orig_dict = json.load(f)
        except Exception as e:
            print(f"{setting_file}: loading failed with error {e}... skipping")
            return False
    if not check_correct_conditions(orig_dict, conditions):
        print(f"{setting_file}: Conditions {conditions} not met ... skipping")
        return False
    as_dict = orig_dict
    try:
        for dict_key in prefixes:
            as_dict = as_dict[dict_key]
    except KeyError:
        print(f"{setting_file}: Not all prefixes {prefixes} present ... skipping")
        return False  # Prefixes not present in JSON file

    if new_key in as_dict:
        if override:
            print(
                f"{setting_file}: Overwrote {'.'.join(prefixes+[new_key])}={default_value} (from {as_dict[new_key]})"
            )
            as_dict[new_key] = default_value
        else:
            print(f"{setting_file}: Key {new_key} already present, not overwriting...")
            return False
    else:
        as_dict[new_key] = default_value
        print(f"{setting_file}: added {'.'.join(prefixes+[new_key])}={default_value}")

    with open(setting_file, "w") as f:
        json.dump(orig_dict, f, indent=4)
    return True



from .file_editing import change_key_name
import argparse
import glob

parser = argparse.ArgumentParser(description="Changing a key name in JSON file(s).")
parser.add_argument(dest="files", type=str, help="Input file(s)")
parser.add_argument(
    dest="key_name",
    type=str,
    help="Name of the key to change (with '.' syntax for nesting)",
)
parser.add_argument(
    dest="new_name",
    type=str,
    help="New name of the key (only the suffix -- no '.' expected)",
)
parser.add_argument(
    "--override", action="store_true", help="Override value if key present"
)


if __name__ == "__main__":
    args = parser.parse_args()

    matching_files = glob.glob(args.files, recursive=True)
    if not matching_files:
        print(f"No files matching {args.files}")
        exit(2)

    *prefixes, old_key = args.key_name.split(".")
    for setting_file in matching_files:
        change_key_name(
            setting_file,
            prefixes=prefixes,
            old_name=old_key,
            new_name=args.new_name,
            conditions=None,
        )



from datetime import datetime
import json
from .param_classes import NoDuplicateDict, recursive_objectify, update_recursive
from .dynamic import recursive_dynamic_json
import collections.abc

try:
    import tomllib
except ImportError:
    # tomllib was added in Python 3.11.  Older versions can use tomli
    import tomli as tomllib
import yaml

from .utils import removesuffix

IMPORT_KEY = "__import__"


def load_raw_dict_from_file(filename):
    """Safe load of a json/yaml/toml file (doubled entries raise exception)."""
    if filename.endswith(".json"):
        with open(filename, "r") as f:
            data = json.load(f, object_pairs_hook=NoDuplicateDict)
    elif filename.endswith(".yaml") or filename.endswith(".yml"):
        with open(filename, "r") as f:
            data = yaml.safe_load(f)
    elif filename.endswith(".toml"):
        with open(filename, "rb") as f:
            data = tomllib.load(f)
    else:
        raise RuntimeError("Unsupported file type.")

    return data


def load(
    filename,
    dynamic=True,
    make_immutable=True,
    recursive_imports=True,
    pre_unpack_hooks=None,
    post_unpack_hooks=None,
):
    """Read from a bytestream and deserialize to a settings object"""
    pre_unpack_hooks = pre_unpack_hooks or []
    post_unpack_hooks = post_unpack_hooks or []
    orig_json = load_raw_dict_from_file(filename)

    for hook in pre_unpack_hooks:
        hook(orig_json)

    if recursive_imports:
        unpack_imports_full(
            orig_json, import_string=IMPORT_KEY, used_filenames=[filename]
        )
    return _post_load(orig_json, dynamic, make_immutable, post_unpack_hooks)


def loads(
    s,
    *,
    dynamic=True,
    make_immutable=False,
    recursive_imports=True,
    pre_unpack_hooks=None,
    post_unpack_hooks=None,
):
    """Deserialize string to a settings object"""
    pre_unpack_hooks = pre_unpack_hooks or []
    post_unpack_hooks = post_unpack_hooks or []

    try:
        orig_dict = json.loads(s, object_pairs_hook=NoDuplicateDict)
    except json.JSONDecodeError as e_json:
        try:
            orig_dict = yaml.safe_load(s)
        except yaml.YAMLError as e_yaml:
            raise SyntaxError(f"JSON and YAML parsing failed: {e_json}, {e_yaml}")

    for hook in pre_unpack_hooks:
        hook(orig_dict)

    if recursive_imports:
        unpack_imports_full(orig_dict, import_string=IMPORT_KEY, used_filenames=[])
    return _post_load(orig_dict, dynamic, make_immutable, post_unpack_hooks)


def _post_load(current_dict, dynamic, make_immutable, post_unpack_hooks):
    keys = list(
        current_dict.keys()
    )  # to avoid that list of keys gets updated during loop
    for key in keys:
        if key.endswith("*") and isinstance(
            current_dict[key], collections.abc.Sequence
        ):
            raw_key = removesuffix(key, "*")
            current_dict[raw_key] = current_dict.pop(key)

    if dynamic:
        objectified = recursive_objectify(current_dict, make_immutable=False)
        timestamp = datetime.now().strftime("%H:%M:%S-%d%h%y")
        namespace = dict(__timestamp__=timestamp, **objectified)
        recursive_dynamic_json(current_dict, namespace)

    for hook in post_unpack_hooks:
        hook(current_dict)

    return recursive_objectify(current_dict, make_immutable=make_immutable)


def dfs_on_dicts(orig_dict):
    for key, value in orig_dict.items():
        if isinstance(value, collections.abc.Mapping):
            yield from dfs_on_dicts(value)
    yield orig_dict


def unpack_imports_full(orig_dict, import_string, used_filenames):
    for current_dict in dfs_on_dicts(orig_dict):
        unpack_imports_fixed_level(current_dict, import_string, used_filenames)


def unpack_imports_fixed_level(orig_dict, import_string, used_filenames):
    """
    If new_files is a list with len(new_files) > 1, items later in the list take
    precedence over files earlier in the list, i.e. keys specified in files later
    in the list override keys specified in files earlier in the list. Keys specified
    in the file that imports the other files have the highest priority and override
    everything else.
    """

    if import_string in orig_dict:
        new_files = orig_dict[
            import_string
        ]  # type(orig_dict[import_string]) in [str, list]
        if isinstance(new_files, str):
            new_files = [new_files]
        del orig_dict[import_string]
        for new_file in reversed(new_files):
            if new_file in used_filenames:
                raise ValueError(
                    f"Cyclic dependency of JSONs, {new_file} already unpacked"
                )
            loaded_dict = load_raw_dict_from_file(new_file)
            unpack_imports_full(loaded_dict, import_string, used_filenames + [new_file])
            update_recursive(orig_dict, loaded_dict, overwrite=False)


if __name__ == "__main__":
    import sys

    def check_import_in_fixed_params(setting_dict):
        if "fixed_params" in setting_dict:
            if "__import__" in setting_dict["fixed_params"]:
                raise ImportError(
                    "Cannot import inside fixed params. Did you mean __import_promise__?"
                )

    params = load(sys.argv[1], pre_unpack_hooks=[check_import_in_fixed_params])
    print(params)



import collections
from copy import deepcopy
import json
from .utils import removesuffix


class NoDuplicateDict(dict):
    """A dict with prohibiting init from a list of pairs containing duplicates"""

    def __init__(self, *args, **kwargs):
        if args and args[0] and not isinstance(args[0], dict):
            keys, _ = zip(*args[0])
            duplicates = [
                item for item, count in collections.Counter(keys).items() if count > 1
            ]
            if duplicates:
                raise TypeError("Keys {} repeated in json parsing".format(duplicates))
        super().__init__(*args, **kwargs)


class AttributeDict(dict):
    """A dict which allows attribute access to its keys."""

    def __getattr__(self, *args, **kwargs):
        try:
            return self.__getitem__(*args, **kwargs)
        except KeyError as e:
            raise AttributeError(e)

    def __deepcopy__(self, memo):
        """In order to support deepcopy"""
        return self.__class__(
            [(deepcopy(k, memo=memo), deepcopy(v, memo=memo)) for k, v in self.items()]
        )

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __delattr__(self, item):
        self.__delitem__(item)

    def __repr__(self):
        return json.dumps(self, indent=4, sort_keys=True)


class ImmutableAttributeDict(AttributeDict):
    """A dict which allows attribute access to its keys. Forced immutable."""

    def __delattr__(self, item):
        raise TypeError("Setting object not mutable after settings are fixed!")

    def __delitem__(self, item):
        raise TypeError("Setting object not mutable after settings are fixed!")

    def __setattr__(self, key, value):
        raise TypeError("Setting object not mutable after settings are fixed!")

    def __setitem__(self, key, value):
        raise TypeError("Setting object not mutable after settings are fixed!")

    def __reduce__(self):
        # For overwriting dict pickling
        # https://stackoverflow.com/questions/21144845/how-can-i-unpickle-a-subclass-of-dict-that-validates-with-setitem-in-pytho
        return (AttributeDict, (self.__getstate__(),))

    def _mutable_copy(self):
        return recursive_objectify(self, make_immutable=False)

    def __getstate__(self):
        return self._mutable_copy()


def recursive_objectify(nested_dict, make_immutable=True):
    "Turns a nested_dict into a nested AttributeDict"
    result = deepcopy(nested_dict)
    for k, v in result.items():
        if isinstance(v, collections.abc.Mapping):
            result = dict(result)
            result[k] = recursive_objectify(v, make_immutable)
    if make_immutable:
        returned_result = ImmutableAttributeDict(result)
    else:
        returned_result = AttributeDict(result)
    return returned_result


def update_recursive(d, u, overwrite=False):
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = update_recursive(d.get(k, {}), v, overwrite)
        if isinstance(v, collections.abc.Sequence):
            raw_key = removesuffix(k, "*")
            if raw_key + "*" in d:  # append
                d[raw_key + "*"] = deepcopy(v + d[raw_key + "*"])
            elif raw_key in d:  # keep original list
                pass
            else:  # key does not exist yet, append
                d[k] = v
        elif k not in d or overwrite:
            d[k] = v
    return d



import os
import collections


def _replace_inside_brackets(string, replace_from, replace_to):
    index = string.find("{")
    if index == -1:
        yield string
        return
    yield string[: index + 1]
    new_index = string[index + 1 :].find("}")
    if new_index == -1:
        yield from string[index:]
        return
    new_index += index + 1
    yield string[index + 1 : new_index].replace(replace_from, replace_to)
    yield from _replace_inside_brackets(string[new_index:], replace_from, replace_to)


def replace_inside_brackets(string, replace_from, replace_to):
    return "".join(list(_replace_inside_brackets(string, replace_from, replace_to)))


def fstring_in_json(format_string, namespace):
    if not isinstance(format_string, str):
        return format_string

    replaced_dollar_signs = replace_inside_brackets(format_string, "$", "ENV_")
    env_dict = {"ENV_" + key: value for key, value in os.environ.items()}
    try:
        formatted = eval('f"' + replaced_dollar_signs + '"', {**env_dict, **namespace})
    except BaseException as e:
        return format_string

    if formatted == format_string:
        return format_string

    try:
        return eval(formatted, dict(__builtins__=None))
    except BaseException as e:
        return formatted


def recursive_dynamic_json(nested_dict_or_list, namespace):
    "Evaluates each key in nested dict as an f-string within a given namespace"
    if isinstance(nested_dict_or_list, collections.abc.Mapping):
        for k, v in nested_dict_or_list.items():
            if isinstance(v, collections.abc.Mapping) or isinstance(v, list):
                recursive_dynamic_json(v, namespace)
            else:
                nested_dict_or_list[k] = fstring_in_json(v, namespace)
    elif isinstance(nested_dict_or_list, list):
        for i, item in enumerate(nested_dict_or_list):
            if isinstance(item, collections.abc.Mapping) or isinstance(item, list):
                recursive_dynamic_json(item, namespace)
            else:
                nested_dict_or_list[i] = fstring_in_json(item, namespace)



from .file_editing import add_key
import argparse
import glob

parser = argparse.ArgumentParser(description="Adding a key to many JSON files.")
parser.add_argument(dest="files", type=str, help="Input file(s)")
parser.add_argument(dest="key_name", type=str, help="Name of the key to add")
parser.add_argument(dest="default_value", type=str, help="Default value to assign")
parser.add_argument(
    "--override", action="store_true", help="Override value if key present"
)


if __name__ == "__main__":
    args = parser.parse_args()

    matching_files = glob.glob(args.files, recursive=True)
    if not matching_files:
        print(f"No files matching {args.files}")
        exit(2)

    *prefixes, new_key = args.key_name.split(".")
    for setting_file in matching_files:
        add_key(
            setting_file,
            prefixes=prefixes,
            new_key=new_key,
            default_value=eval(args.default_value),
            override=args.override,
            conditions=None,
        )
